from typing import (Callable,
                    TypeVar)

from typing_extensions import final

from lz._core.functional import ApplierBase
from lz._core.signatures import (Signature,
                                 to_signature)

_Arg = TypeVar('_Arg')
_KwArg = TypeVar('_KwArg')
_Result = TypeVar('_Result')


@final
class Applier(ApplierBase[_Arg, _KwArg, _Result]):
    def __init__(self,
                 function: Callable[..., _Result],
                 *args: _Arg,
                 **kwargs: _KwArg) -> None:
        super().__init__(function, *args, **kwargs)

    def __call__(self, *args: _Arg, **kwargs: _KwArg) -> _Result:
        return self.function(*self.args, *args, **self.kwargs, **kwargs)


@to_signature.register(Applier)
def _(_value: Applier[_Arg, _KwArg, _Result]) -> Signature:
    return to_signature(_value.function).bind(*_value.args, **_value.kwargs)



from __future__ import annotations

import ast
import functools
import itertools
import sys
import types
import typing as _t
from abc import (ABC,
                 abstractmethod)
from types import MethodType

import typing_extensions as _te
from paradigm.base import (OverloadedSignature,
                           PlainSignature)
from reprit import seekers
from reprit.base import generate_repr

from lz._core.signatures import (Signature,
                                 to_signature)

_Arg = _t.TypeVar('_Arg')
_KwArg = _t.TypeVar('_KwArg')
_Result = _t.TypeVar('_Result')
_T = _t.TypeVar('_T')

MIN_COMPOSABLE_FUNCTIONS_COUNT = 2


@_te.final
class Composition(_t.Generic[_Arg, _KwArg, _Result]):
    _file_path: str
    _function: _t.Callable[..., _Result]
    _functions: _t.Tuple[_t.Callable[..., _t.Any], ...]
    _line_number: int
    _line_offset: int

    __slots__ = ('_file_path', '_function', '_functions', '_line_number',
                 '_line_offset')

    def __new__(cls,
                *functions: _t.Callable[..., _t.Any],
                file_path: str = __file__,
                line_number: int = 0,
                line_offset: int = 0) -> Composition[_Arg, _KwArg, _Result]:
        if len(functions) < MIN_COMPOSABLE_FUNCTIONS_COUNT:
            raise ValueError('Composition is defined '
                             f'for {MIN_COMPOSABLE_FUNCTIONS_COUNT} '
                             'or more functions, '
                             f'but got {len(functions)}.')
        self = super().__new__(cls)

        def flatten(
                function: _t.Callable[..., _t.Any]
        ) -> _t.Iterable[_t.Callable[..., _t.Any]]:
            if isinstance(function, cls):
                yield from function._functions
            else:
                yield function

        self._functions = tuple(itertools.chain.from_iterable(map(flatten,
                                                                  functions)))
        self._file_path = file_path
        self._line_number = line_number
        self._line_offset = line_offset
        self._function = _compose(*functions,
                                  function_name='composition',
                                  file_path=file_path,
                                  line_number=line_number,
                                  line_offset=line_offset)
        return self

    def __call__(self, *args: _Arg, **kwargs: _KwArg) -> _Result:
        return self._function(*args, **kwargs)

    def __get__(self,
                instance: _T,
                owner: _t.Type[_T]) -> _t.Callable[..., _Result]:
        return MethodType(self, instance)

    def __getnewargs_ex__(self) -> _t.Tuple[
        _t.Tuple[_t.Any, ...], _t.Dict[str, _t.Any]
    ]:
        return self._functions, {'file_path': self._file_path,
                                 'line_number': self._line_number,
                                 'line_offset': self._line_offset}

    def __getstate__(self) -> None:
        return None

    def __setstate__(self, _state: None) -> None:
        pass

    __repr__ = generate_repr(__new__,
                             field_seeker=seekers.complex_)


@_te.final
class Combination(_t.Generic[_Arg, _Result]):
    _file_path: str
    _function: _t.Callable[..., _t.Tuple[_Result, ...]]
    _line_number: int
    _line_offset: int
    _maps: _t.Tuple[_t.Callable[[_Arg], _Result], ...]

    __slots__ = ('_file_path', '_function', '_line_number', '_line_offset',
                 '_maps')

    def __new__(cls,
                *_maps: _t.Callable[[_Arg], _Result],
                file_path: str = __file__,
                line_number: int = 0,
                line_offset: int = 0) -> Combination[_Arg, _Result]:
        self = super().__new__(cls)
        self._maps = _maps
        self._file_path = file_path
        self._line_number = line_number
        self._line_offset = line_offset
        self._function = _combine(*_maps,
                                  function_name='combination',
                                  file_path=file_path,
                                  line_number=line_number,
                                  line_offset=line_offset)
        return self

    def __call__(self, *args: _Arg) -> _t.Tuple[_Result, ...]:
        return self._function(*args)

    def __getnewargs_ex__(self) -> _t.Tuple[
        _t.Tuple[_t.Any, ...], _t.Dict[str, _t.Any]
    ]:
        return self._maps, {'file_path': self._file_path,
                            'line_number': self._line_number,
                            'line_offset': self._line_offset}

    def __getstate__(self) -> None:
        return None

    def __setstate__(self, _state: None) -> None:
        pass

    __repr__ = generate_repr(__new__,
                             field_seeker=seekers.complex_)


class ApplierBase(ABC, _t.Generic[_Arg, _KwArg, _Result]):
    def __init__(self,
                 function: _t.Callable[..., _Result],
                 *args: _Arg,
                 **kwargs: _KwArg) -> None:
        if isinstance(function, type(self)):
            args = function.args + args
            kwargs = {**function.kwargs, **kwargs}
            function = function.function
        self._function = function
        self._args = args
        self._kwargs = kwargs

    @property
    def function(self) -> _t.Callable[..., _Result]:
        return self._function

    @property
    def args(self) -> _t.Tuple[_Arg, ...]:
        return self._args

    @property
    def kwargs(self) -> _t.Dict[str, _KwArg]:
        return self._kwargs

    @abstractmethod
    def __call__(self, *args: _Arg, **kwargs: _KwArg) -> _t.Any:
        pass


@_te.final
class Curry(ApplierBase[_Arg, _KwArg, _Result]):
    def __init__(self,
                 function: _t.Callable[..., _Result],
                 _signature: Signature,
                 *args: _Arg,
                 **kwargs: _KwArg) -> None:
        super().__init__(function, *args, **kwargs)
        self._signature = _signature

    def __call__(
            self, *args: _Arg, **kwargs: _KwArg
    ) -> _t.Union[Curry[_Arg, _KwArg, _Result], _Result]:
        total_args = self.args + args
        total_kwargs = {**self.kwargs, **kwargs}
        try:
            return self.function(*total_args, **total_kwargs)
        except TypeError:
            if (not self._signature.expects(*total_args, **total_kwargs)
                    or self._signature.all_set(*total_args, **total_kwargs)):
                raise
            return type(self)(self.function, self._signature, *total_args,
                              **total_kwargs)

    def __getstate__(self) -> _t.Tuple[
        _t.Callable[..., _Result], _t.Tuple[_Arg, ...], _t.Dict[str, _KwArg]
    ]:
        return self.function, self.args, self.kwargs

    def __setstate__(self,
                     state: _t.Tuple[
                         _t.Callable[..., _Result], _t.Tuple[_Arg, ...],
                         _t.Dict[str, _KwArg]
                     ]) -> None:
        self._function, self._args, self._kwargs = state
        self._signature = to_signature(self._function)

    __repr__ = generate_repr(__init__,
                             field_seeker=seekers.complex_)


@_te.final
class Constant(_t.Generic[_T]):
    def __init__(self, _value: _T) -> None:
        self._value = _value

    def __call__(self, *args: _t.Any, **kwargs: _t.Any) -> _T:
        return self._value

    __repr__ = generate_repr(__init__)


_Params = _te.ParamSpec('_Params')

if sys.version_info < (3, 10):
    @_te.final
    class Cleavage(_t.Generic[_Result]):
        _file_path: str
        _function: _t.Callable[..., _t.Tuple[_Result, ...]]
        _functions: _t.Tuple[_t.Callable[..., _Result], ...]
        _line_number: int
        _line_offset: int

        __slots__ = ('_file_path', '_function', '_functions', '_line_number',
                     '_line_offset')

        def __new__(cls,
                    *functions: _t.Callable[_Params, _Result],
                    file_path: str = __file__,
                    line_number: int = 0,
                    line_offset: int = 0) -> _te.Self:
            self = super().__new__(cls)
            self._functions = functions
            self._file_path = file_path
            self._line_number = line_number
            self._line_offset = line_offset
            self._function = _cleave(*functions,
                                     function_name='cleavage',
                                     file_path=file_path,
                                     line_number=line_number,
                                     line_offset=line_offset)
            return self

        def __call__(self,
                     *args: _Params.args,
                     **kwargs: _Params.kwargs) -> _t.Tuple[_Result, ...]:
            return self._function(*args, **kwargs)

        def __getnewargs_ex__(self) -> _t.Tuple[
            _t.Tuple[_t.Any, ...], _t.Dict[str, _t.Any]
        ]:
            return self._functions, {'file_path': self._file_path,
                                     'line_number': self._line_number,
                                     'line_offset': self._line_offset}

        def __getstate__(self) -> None:
            return None

        def __setstate__(self, _state: None) -> None:
            pass

        __repr__ = generate_repr(__new__,
                                 field_seeker=seekers.complex_)
else:
    @_te.final
    class Cleavage(_t.Generic[_Params, _Result]):
        _file_path: str
        _function: _t.Callable[_Params, _t.Tuple[_Result, ...]]
        _functions: _t.Tuple[_t.Callable[_Params, _Result], ...]
        _line_number: int
        _line_offset: int

        __slots__ = ('_file_path', '_function', '_functions', '_line_number',
                     '_line_offset')

        def __new__(cls,
                    *functions: _t.Callable[_Params, _Result],
                    file_path: str = __file__,
                    line_number: int = 0,
                    line_offset: int = 0) -> _te.Self:
            self = super().__new__(cls)
            self._functions = functions
            self._file_path = file_path
            self._line_number = line_number
            self._line_offset = line_offset
            self._function = _cleave(*functions,
                                     function_name='cleavage',
                                     file_path=file_path,
                                     line_number=line_number,
                                     line_offset=line_offset)
            return self

        def __call__(self,
                     *args: _Params.args,
                     **kwargs: _Params.kwargs) -> _t.Tuple[_Result, ...]:
            return self._function(*args, **kwargs)

        def __getnewargs_ex__(self) -> _t.Tuple[
            _t.Tuple[_t.Any, ...], _t.Dict[str, _t.Any]
        ]:
            return self._functions, {'file_path': self._file_path,
                                     'line_number': self._line_number,
                                     'line_offset': self._line_offset}

        def __getstate__(self) -> None:
            return None

        def __setstate__(self, _state: None) -> None:
            pass

        __repr__ = generate_repr(__new__,
                                 field_seeker=seekers.complex_)


@_te.final
class Flip(_t.Generic[_Result]):
    if sys.version_info < (3, 10):
        @classmethod
        @_t.overload
        def from_function(
                cls, _function: Cleavage[_Result]
        ) -> Cleavage[_Result]:
            ...
    else:
        @classmethod
        @_t.overload
        def from_function(
                cls, _function: Cleavage[_Params, _Result]
        ) -> Cleavage[_Params, _Result]:
            ...

    @classmethod
    @_t.overload
    def from_function(
            cls, _function: Composition[_Arg, _KwArg, _Result]
    ) -> Composition[_Arg, _KwArg, _Result]:
        ...

    @classmethod
    @_t.overload
    def from_function(cls, _function: Constant[_Result]) -> Constant[_Result]:
        ...

    @classmethod
    @_t.overload
    def from_function(
            cls, _function: Flip[_Result]
    ) -> _t.Callable[_Params, _Result]:
        ...

    @classmethod
    @_t.overload
    def from_function(
            cls, _function: _t.Callable[_Params, _Result]
    ) -> _t.Callable[_Params, _Result]:
        ...

    @classmethod
    def from_function(cls, _function: _t.Any) -> _t.Any:
        return (_function._function
                if isinstance(_function, cls)
                else (Composition(*_function._functions[:-1],
                                  cls.from_function(_function._functions[-1]),
                                  file_path=_function._file_path,
                                  line_number=_function._line_number,
                                  line_offset=_function._line_offset)
                      if isinstance(_function, Composition)
                      else (Cleavage(*[cls.from_function(function)
                                       for function in _function._functions],
                                     file_path=_function._file_path,
                                     line_number=_function._line_number,
                                     line_offset=_function._line_offset)
                            if isinstance(_function, Cleavage)
                            else (_function
                                  if isinstance(_function, Constant)
                                  else cls(_function)))))

    _function: _t.Callable[..., _Result]

    __slots__ = '_function',

    def __new__(cls, _function: _t.Callable[..., _Result]) -> Flip[_Result]:
        if isinstance(_function, cls):
            raise ValueError('Repeated flip should return original function.')
        self = super().__new__(cls)
        self._function = _function
        return self

    def __call__(self, *args: _Arg, **kwargs: _KwArg) -> _Result:
        return self._function(*args[::-1], **kwargs)

    def __getnewargs__(self) -> _t.Tuple[_t.Any, ...]:
        return (self._function,)

    __repr__ = generate_repr(__new__)


def _cleave(
        *functions: _t.Callable[_Params, _Result],
        function_name: str,
        file_path: str,
        line_number: int,
        line_offset: int,
        args_name: str = 'args',
        kwargs_name: str = 'kwargs'
) -> _t.Callable[_Params, _t.Tuple[_Result, ...]]:
    functions_names = [_function_to_unique_name(function)
                       for function in functions]
    result_node = ast.Tuple(
            [
                ast.Call(ast.Name(function_name, ast.Load(),
                                  lineno=line_number,
                                  col_offset=line_offset),
                         [ast.Starred(ast.Name(args_name, ast.Load(),
                                               lineno=line_number,
                                               col_offset=line_offset),
                                      ast.Load(),
                                      lineno=line_number,
                                      col_offset=line_offset)],
                         [ast.keyword(None, ast.Name(kwargs_name, ast.Load(),
                                                     lineno=line_number,
                                                     col_offset=line_offset),
                                      lineno=line_number,
                                      col_offset=line_offset)],
                         lineno=line_number,
                         col_offset=line_offset)
                for function_name in functions_names
            ],
            ast.Load(),
            lineno=line_number,
            col_offset=line_offset
    )
    function_definition_node = ast.FunctionDef(
            function_name,
            _to_signature_node(
                    variadic_positional_parameter=ast.arg(
                            args_name, None,
                            lineno=line_number,
                            col_offset=line_offset
                    ),
                    variadic_keyword_parameter=ast.arg(kwargs_name, None,
                                                       lineno=line_number,
                                                       col_offset=line_offset)
            ),
            [ast.Return(result_node,
                        lineno=line_number,
                        col_offset=line_offset)],
            [], None,
            lineno=line_number,
            col_offset=line_offset
    )
    return _compile_function(function_definition_node,
                             file_path=file_path,
                             namespace=dict(zip(functions_names, functions)))


def _combine(*maps: _t.Callable[[_T], _Result],
             function_name: str,
             file_path: str,
             line_number: int,
             line_offset: int) -> _t.Callable[..., _t.Tuple[_Result, ...]]:
    maps_names = [_function_to_unique_name(map_) for map_ in maps]
    args_names = [f'_arg{index}' for index in range(len(maps))]
    function_definition_node = ast.FunctionDef(
            function_name,
            _to_signature_node([ast.arg(arg_name, None,
                                        lineno=line_number,
                                        col_offset=line_offset)
                                for arg_name in args_names]),
            [ast.Return(ast.Tuple([ast.Call(ast.Name(map_name, ast.Load(),
                                                     lineno=line_number,
                                                     col_offset=line_offset),
                                            [ast.Name(arg_name, ast.Load(),
                                                      lineno=line_number,
                                                      col_offset=line_offset)],
                                            [],
                                            lineno=line_number,
                                            col_offset=line_offset)
                                   for map_name, arg_name in zip(maps_names,
                                                                 args_names)],
                                  ast.Load(),
                                  lineno=line_number,
                                  col_offset=line_offset),
                        lineno=line_number,
                        col_offset=line_offset)],
            [], None,
            lineno=line_number,
            col_offset=line_offset
    )
    return _compile_function(function_definition_node,
                             file_path=file_path,
                             namespace=dict(zip(maps_names, maps)))


def _compose(*functions: _t.Callable[..., _t.Any],
             function_name: str,
             file_path: str,
             line_number: int,
             line_offset: int,
             args_name: str = 'args',
             kwargs_name: str = 'kwargs') -> _t.Callable[..., _Result]:
    functions_names = [_function_to_unique_name(function)
                       for function in functions]

    def to_next_call_node(node: ast.Call, name: str) -> ast.Call:
        return ast.Call(to_name_node(name), [node], [],
                        lineno=line_number,
                        col_offset=line_offset)

    def to_name_node(name: str,
                     *,
                     context_factory: _t.Type[ast.expr_context] = ast.Load
                     ) -> ast.Name:
        return ast.Name(name, context_factory(),
                        lineno=line_number,
                        col_offset=line_offset)

    reversed_functions_names = reversed(functions_names)
    calls_node = ast.Call(to_name_node(next(reversed_functions_names)),
                          [ast.Starred(to_name_node(args_name), ast.Load(),
                                       lineno=line_number,
                                       col_offset=line_offset)],
                          [ast.keyword(None, to_name_node(kwargs_name),
                                       lineno=line_number,
                                       col_offset=line_offset)],
                          lineno=line_number,
                          col_offset=line_offset)
    calls_node = functools.reduce(to_next_call_node,
                                  reversed_functions_names,
                                  calls_node)
    function_definition_node = ast.FunctionDef(
            function_name,
            _to_signature_node(
                    variadic_positional_parameter=ast.arg(
                            args_name, None,
                            lineno=line_number,
                            col_offset=line_offset
                    ),
                    variadic_keyword_parameter=ast.arg(kwargs_name, None,
                                                       lineno=line_number,
                                                       col_offset=line_offset)
            ),
            [ast.Return(calls_node,
                        lineno=line_number,
                        col_offset=line_offset)],
            [],
            None,
            lineno=line_number,
            col_offset=line_offset
    )
    return _compile_function(function_definition_node,
                             file_path=file_path,
                             namespace=dict(zip(functions_names, functions)))


def _compile_function(
        function_definition_node: ast.FunctionDef,
        *,
        file_path: str,
        namespace: _t.Dict[str, _t.Any],
        module_factory: _t.Callable[..., ast.Module] =
        ast.Module
        if sys.version_info < (3, 8)
        # Python3.8 adds `type_ignores` parameter
        else _t.cast(_t.Callable[..., ast.Module],
                     functools.partial(ast.Module,
                                       type_ignores=[]))
) -> _t.Callable[..., _t.Any]:
    tree = module_factory([function_definition_node])
    code = compile(tree, file_path, 'exec')
    exec(code, namespace)
    result = namespace[function_definition_node.name]
    assert isinstance(result, types.FunctionType)
    return result


if sys.version_info < (3, 8):
    def _to_signature_node(
            positional_only_parameters: _t.Optional[_t.List[ast.arg]] = None,
            positional_or_keyword_parameters: _t.Optional[_t.List[ast.arg]]
            = None,
            variadic_positional_parameter: _t.Optional[ast.arg] = None,
            keyword_only_parameters: _t.Optional[_t.List[ast.arg]] = None,
            variadic_keyword_parameter: _t.Optional[ast.arg] = None,
            positionals_defaults: _t.Optional[_t.List[ast.expr]] = None,
            keywords_defaults: _t.Optional[_t.List[_t.Optional[ast.expr]]]
            = None
    ) -> ast.arguments:
        return ast.arguments((positional_only_parameters or [])
                             + (positional_or_keyword_parameters or []),
                             variadic_positional_parameter,
                             keyword_only_parameters or [],
                             keywords_defaults or [],
                             variadic_keyword_parameter,
                             positionals_defaults or [])
else:
    def _to_signature_node(
            positional_only_parameters: _t.Optional[_t.List[ast.arg]] = None,
            positional_or_keyword_parameters: _t.Optional[_t.List[ast.arg]]
            = None,
            variadic_positional_parameter: _t.Optional[ast.arg] = None,
            keyword_only_parameters: _t.Optional[_t.List[ast.arg]] = None,
            variadic_keyword_parameter: _t.Optional[ast.arg] = None,
            positionals_defaults: _t.Optional[_t.List[ast.expr]] = None,
            keywords_defaults: _t.Optional[_t.List[_t.Optional[ast.expr]]]
            = None
    ) -> ast.arguments:
        return ast.arguments(positional_only_parameters or [],
                             positional_or_keyword_parameters or [],
                             variadic_positional_parameter,
                             keyword_only_parameters or [],
                             keywords_defaults or [],
                             variadic_keyword_parameter,
                             positionals_defaults or [])


def _function_to_unique_name(function: _t.Callable[..., _t.Any]) -> str:
    # we are not using `__name__`/`__qualname__` attributes
    # due to their potential non-uniqueness
    return '_' + str(id(function)).replace('-', '_')


@to_signature.register(Combination)
def _(_value: Combination[_Arg, _Result]) -> Signature:
    return to_signature(_value._function)


@to_signature.register(Composition)
def _(_value: Composition[_Arg, _KwArg, _Result]) -> Signature:
    last_signature = to_signature(_value._functions[0])
    returns = (_t.Union[tuple(signature.returns
                              for signature in last_signature.signatures
                              if signature.expects(None))]
               if isinstance(last_signature, OverloadedSignature)
               else last_signature.returns)
    return _replace_returns(to_signature(_value._functions[-1]), returns)


if sys.version_info < (3, 10):
    @to_signature.register(Cleavage)
    def _(_value: Cleavage[_Result]) -> Signature:
        return to_signature(_value._functions[0])
else:
    @to_signature.register(Cleavage)
    def _(_value: Cleavage[_Params, _Result]) -> Signature:
        return to_signature(_value._functions[0])


@functools.singledispatch
def _replace_returns(signature: Signature, returns: _t.Any) -> Signature:
    raise TypeError(type(signature))


@_replace_returns.register(OverloadedSignature)
def _(signature: OverloadedSignature, returns: _t.Any) -> OverloadedSignature:
    return OverloadedSignature(
            *[_replace_returns_for_plain_signature(sub_signature, returns)
              for sub_signature in signature.signatures]
    )


@_replace_returns.register(PlainSignature)
def _replace_returns_for_plain_signature(signature: PlainSignature,
                                         returns: _t.Any) -> PlainSignature:
    return PlainSignature(*signature.parameters,
                          returns=returns)


@to_signature.register(Constant)
def _(_value: Constant[_T]) -> Signature:
    return to_signature(_value.__call__)


@to_signature.register(Curry)
def _(_value: Curry[_Arg, _KwArg, _Result]) -> Signature:
    return _value._signature


@to_signature.register(Flip)
def _(_value: Flip[_Result]) -> Signature:
    return to_signature(_value._function)



import codecs as _codecs
import functools as _functools
import io as _io
import itertools as _itertools
import operator as _operator
import typing as _t
from collections import abc as _abc

from lz._core.arithmetical import ceil_division as _ceil_division
from lz._core.textual import (code_units_sizes as _code_units_sizes,
                              read_batch_from_end as _read_batch_from_end)

_T = _t.TypeVar('_T')


@_functools.singledispatch
def reverse(_value: _t.Any) -> _t.Any:
    """
    Returns reversed object.

    >>> list(reverse(range(10)))
    [9, 8, 7, 6, 5, 4, 3, 2, 1, 0]
    >>> import io
    >>> list(reverse(_io.BytesIO(b'Hello\\nWorld!')))
    [b'World!', b'Hello\\n']
    """
    raise TypeError('Unsupported object type: {type}.'
                    .format(type=type(_value)))


@reverse.register(_abc.Sequence)
def _(_value: _t.Sequence[_T]) -> _t.Sequence[_T]:
    """
    Returns reversed sequence.
    """
    return _value[::-1]


@reverse.register(_abc.Reversible)
def _(_value: _t.Reversible[_T]) -> _t.Iterable[_T]:
    """
    Returns reversed reversible iterable.
    """
    yield from reversed(_value)


@reverse.register(_io.TextIOWrapper)
def reverse_file_object(_value: _t.TextIO,
                        *,
                        batch_size: int = _io.DEFAULT_BUFFER_SIZE,
                        lines_separator: _t.Optional[str] = None,
                        keep_lines_separator: bool = True) -> _t.Iterable[str]:
    """
    Returns reversed file object.
    """
    encoding = _value.encoding
    code_unit_size = _code_units_sizes[encoding]
    bytes_lines_separator = (None
                             if lines_separator is None
                             else lines_separator.encode(encoding))
    yield from map(
            _functools.partial(_codecs.decode,
                               encoding=encoding),
            reverse_bytes_stream(_value.buffer,
                                 batch_size=batch_size,
                                 lines_separator=bytes_lines_separator,
                                 keep_lines_separator=keep_lines_separator,
                                 code_unit_size=code_unit_size)
    )


@reverse.register(_io.BufferedReader)
@reverse.register(_io.BytesIO)
def reverse_bytes_stream(_value: _t.BinaryIO,
                         *,
                         batch_size: int = _io.DEFAULT_BUFFER_SIZE,
                         lines_separator: _t.Optional[bytes] = None,
                         keep_lines_separator: bool = True,
                         code_unit_size: int = 1) -> _t.Iterable[bytes]:
    """
    Returns reversed byte stream.
    """
    lines_separators: _t.Union[bytes, _t.Tuple[bytes, ...]]
    lines_splitter: _t.Callable[[bytes], _t.List[bytes]]
    if lines_separator is None:
        lines_separators = (b'\r', b'\n', b'\r\n')
        lines_splitter = _operator.methodcaller(bytes.splitlines.__name__,
                                                keep_lines_separator)
    else:
        lines_separators = lines_separator
        _separator = lines_separator

        def lines_splitter(byte_sequence: bytes,
                           separator: bytes = _separator) -> _t.List[bytes]:
            result: _t.List[bytes] = []
            part = bytearray()
            offset = 0
            add_part = result.append
            while offset < len(byte_sequence):
                if (byte_sequence[offset:offset + len(separator)]
                        != separator):
                    part += byte_sequence[offset:offset + code_unit_size]
                    offset += code_unit_size
                else:
                    add_part(part + keep_lines_separator * separator)
                    part.clear()
                    offset += len(separator)
            add_part(part)
            return result
    stream_size = _value.seek(0, _io.SEEK_END)
    batches_count = _ceil_division(stream_size, batch_size)
    remaining_bytes_indicator = _itertools.islice(
            _itertools.accumulate(
                    _itertools.chain([stream_size],
                                     _itertools.repeat(batch_size)),
                    _operator.sub
            ),
            batches_count
    )
    try:
        remaining_bytes_count = next(remaining_bytes_indicator)
    except StopIteration:
        return

    def read_batch(position: int) -> bytes:
        result = _read_batch_from_end(_value,
                                      size=batch_size,
                                      end_position=position)
        while result.startswith(lines_separators):
            try:
                position = next(remaining_bytes_indicator)
            except StopIteration:
                break
            result = (_read_batch_from_end(_value,
                                           size=batch_size,
                                           end_position=position)
                      + result)
        return result

    batch = read_batch(remaining_bytes_count)
    segment, *lines = lines_splitter(batch)
    yield from reverse(lines)
    for remaining_bytes_count in remaining_bytes_indicator:
        batch = read_batch(remaining_bytes_count)
        lines = lines_splitter(batch)
        if batch.endswith(lines_separators):
            yield segment
        else:
            lines[-1] += segment
        segment, *lines = lines
        yield from reverse(lines)
    yield segment



import functools as _functools
import itertools as _itertools
import typing as _t
from collections import deque as _deque

from .replication import duplicate as _duplicate

_T = _t.TypeVar('_T')


def sift(_predicate: _t.Callable[[_T], bool],
         _value: _t.Iterable[_T]) -> _t.Iterable[_T]:
    """
    Selects elements from iterable which satisfy given predicate.

    >>> list(sift(bool, range(10)))
    [1, 2, 3, 4, 5, 6, 7, 8, 9]

    >>> def is_even(number: int) -> bool:
    ...     return number % 2 == 0
    >>> list(sift(is_even, range(10)))
    [0, 2, 4, 6, 8]
    """
    return filter(_predicate, _value)


def sifter(
        _predicate: _t.Callable[[_T], bool]
) -> _t.Callable[[_t.Iterable[_T]], _t.Iterable[_T]]:
    """
    Returns function that selects elements from iterable
    which satisfy given predicate.

    >>> to_true_like = sifter(bool)
    >>> list(to_true_like(range(10)))
    [1, 2, 3, 4, 5, 6, 7, 8, 9]

    >>> def is_even(number: int) -> bool:
    ...     return number % 2 == 0
    >>> to_even = sifter(is_even)
    >>> list(to_even(range(10)))
    [0, 2, 4, 6, 8]
    """
    return _functools.partial(sift, _predicate)


def scavenge(_predicate: _t.Callable[[_T], bool],
             _value: _t.Iterable[_T]) -> _t.Iterable[_T]:
    """
    Selects elements from iterable which dissatisfy given predicate.

    >>> list(scavenge(bool, range(10)))
    [0]

    >>> def is_even(number: int) -> bool:
    ...     return number % 2 == 0
    >>> list(scavenge(is_even, range(10)))
    [1, 3, 5, 7, 9]
    """
    return _itertools.filterfalse(_predicate, _value)


def scavenger(
        _predicate: _t.Callable[[_T], bool]
) -> _t.Callable[[_t.Iterable[_T]], _t.Iterable[_T]]:
    """
    Returns function that selects elements from iterable
    which dissatisfy given predicate.

    >>> to_false_like = scavenger(bool)
    >>> list(to_false_like(range(10)))
    [0]

    >>> def is_even(number: int) -> bool:
    ...     return number % 2 == 0
    >>> to_odd = scavenger(is_even)
    >>> list(to_odd(range(10)))
    [1, 3, 5, 7, 9]
    """
    return _functools.partial(scavenge, _predicate)


def separate(
        _predicate: _t.Callable[[_T], bool], _value: _t.Iterable[_T]
) -> _t.Tuple[_t.Iterable[_T], _t.Iterable[_T]]:
    """
    Returns pair of iterables
    first of which consists of elements that dissatisfy given predicate
    and second one consists of elements that satisfy given predicate.

    >>> tuple(map(list, separate(bool, range(10))))
    ([0], [1, 2, 3, 4, 5, 6, 7, 8, 9])

    >>> def is_even(number: int) -> bool:
    ...     return number % 2 == 0
    >>> tuple(map(list, separate(is_even, range(10))))
    ([1, 3, 5, 7, 9], [0, 2, 4, 6, 8])
    """
    iterator = iter(_value)
    unsatisfying: _t.Deque[_T] = _deque()
    satisfying: _t.Deque[_T] = _deque()

    def fill(queue: _t.Deque[_T]) -> _t.Iterable[_T]:
        while True:
            while not queue:
                try:
                    element = next(iterator)
                except StopIteration:
                    return
                subject, element = _duplicate(element)
                (satisfying
                 if _predicate(subject)
                 else unsatisfying).append(element)
            yield queue.popleft()

    return fill(unsatisfying), fill(satisfying)


def separator(
        _predicate: _t.Callable[[_T], bool]
) -> _t.Callable[[_t.Iterable[_T]],
                 _t.Tuple[_t.Iterable[_T], _t.Iterable[_T]]]:
    """
    Returns function that returns pair of iterables
    first of which consists of elements that dissatisfy given predicate
    and second one consists of elements that satisfy given predicate.

    >>> split_by_truth = separator(bool)
    >>> tuple(map(list, split_by_truth(range(10))))
    ([0], [1, 2, 3, 4, 5, 6, 7, 8, 9])

    >>> def is_even(number: int) -> bool:
    ...     return number % 2 == 0
    >>> split_by_evenness = separator(is_even)
    >>> tuple(map(list, split_by_evenness(range(10))))
    ([1, 3, 5, 7, 9], [0, 2, 4, 6, 8])
    """
    return _functools.partial(separate, _predicate)


def grab(_predicate: _t.Callable[[_T], bool],
         _value: _t.Iterable[_T]) -> _t.Iterable[_T]:
    """
    Selects elements from the beginning of iterable
    while given predicate is satisfied.

    >>> grab_while_true_like = grabber(bool)
    >>> list(grab_while_true_like(range(10)))
    []

    >>> from operator import gt
    >>> from functools import partial
    >>> grab_while_less_than_five = grabber(partial(gt, 5))
    >>> list(grab_while_less_than_five(range(10)))
    [0, 1, 2, 3, 4]
    """
    return _itertools.takewhile(_predicate, _value)


def grabber(
        _predicate: _t.Callable[[_T], bool]
) -> _t.Callable[[_t.Iterable[_T]], _t.Iterable[_T]]:
    """
    Returns function that selects elements from the beginning of iterable
    while given predicate is satisfied.

    >>> grab_while_true_like = grabber(bool)
    >>> list(grab_while_true_like(range(10)))
    []

    >>> from operator import gt
    >>> from functools import partial
    >>> grab_while_less_than_five = grabber(partial(gt, 5))
    >>> list(grab_while_less_than_five(range(10)))
    [0, 1, 2, 3, 4]
    """
    return _functools.partial(grab, _predicate)


def kick(_predicate: _t.Callable[[_T], bool],
         _value: _t.Iterable[_T]) -> _t.Iterable[_T]:
    """
    Skips elements from the beginning of iterable
    while given predicate is satisfied.

    >>> list(kick(bool, range(10)))
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

    >>> from operator import gt
    >>> from functools import partial
    >>> list(kick(partial(gt, 5), range(10)))
    [5, 6, 7, 8, 9]
    """
    return _itertools.dropwhile(_predicate, _value)


def kicker(
        _predicate: _t.Callable[[_T], bool]
) -> _t.Callable[[_t.Iterable[_T]], _t.Iterable[_T]]:
    """
    Returns function that skips elements from the beginning of iterable
    while given predicate is satisfied.

    >>> kick_while_true_like = kicker(bool)
    >>> list(kick_while_true_like(range(10)))
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

    >>> from operator import gt
    >>> from functools import partial
    >>> kick_while_less_than_five = kicker(partial(gt, 5))
    >>> list(kick_while_less_than_five(range(10)))
    [5, 6, 7, 8, 9]
    """
    return _functools.partial(kick, _predicate)



import functools as _functools
import itertools as _itertools
import typing as _t
from collections import (abc as _abc,
                         deque as _deque)

_T = _t.TypeVar('_T')


@_functools.singledispatch
def replicate(_value: _t.Any,
              *,
              count: int) -> _t.Iterable[_t.Any]:
    """
    Returns given number of object replicas.
    """
    raise TypeError('Unsupported object type: {type}.'
                    .format(type=type(_value)))


_Immutable = _t.TypeVar('_Immutable', bytes, object, str)


@replicate.register(object)
# immutable strings represent a special kind of iterables
# that can be replicated by simply repeating
@replicate.register(bytes)
@replicate.register(str)
def _(_value: _Immutable, *, count: int) -> _t.Iterable[_Immutable]:
    """
    Returns object repeated given number of times.
    """
    yield from _itertools.repeat(_value, count)


@replicate.register(_abc.Iterable)
def _replicate_iterable(_value: _t.Iterable[_T],
                        *,
                        count: int) -> _t.Iterable[_t.Iterable[_T]]:
    """
    Returns given number of iterable replicas.
    """
    iterator = iter(_value)
    queues: _t.Sequence[_t.Deque[_T]] = [
        _deque() for _ in _itertools.repeat(None, count)
    ]

    def replica(queue: _t.Deque[_T]) -> _t.Iterable[_T]:
        while True:
            if not queue:
                try:
                    element = next(iterator)
                except StopIteration:
                    return
                element_copies = replicate(element,
                                           count=count)
                for sub_queue, element_copy in zip(queues, element_copies):
                    sub_queue.append(element_copy)
            yield queue.popleft()

    yield from map(replica, queues)


@replicate.register(bytearray)
def _(_value: bytearray,
      *,
      count: int) -> _t.Iterable[bytearray]:
    for _ in _itertools.repeat(None, count):
        yield _value[:]


@replicate.register(frozenset)
def _(_value: _t.FrozenSet[_T],
      *,
      count: int) -> _t.Iterable[_t.FrozenSet[_T]]:
    for replica in _replicate_iterable(_value,
                                       count=count):
        yield frozenset(replica)


@replicate.register(list)
def _(_value: _t.List[_T],
      *,
      count: int) -> _t.Iterable[_t.List[_T]]:
    for replica in _replicate_iterable(_value,
                                       count=count):
        yield list(replica)


@replicate.register(set)
def _(_value: _t.Set[_T],
      *,
      count: int) -> _t.Iterable[_t.Set[_T]]:
    for replica in _replicate_iterable(_value,
                                       count=count):
        yield set(replica)


@replicate.register(tuple)
def _(_value: _t.Tuple[_T, ...],
      *,
      count: int) -> _t.Iterable[_t.Tuple[_T, ...]]:
    for replica in _replicate_iterable(_value,
                                       count=count):
        yield tuple(replica)


_Key = _t.TypeVar('_Key')
_Value = _t.TypeVar('_Value')


@replicate.register(dict)
def _(_value: _t.Dict[_Key, _Value],
      *,
      count: int) -> _t.Iterable[_t.Dict[_Key, _Value]]:
    for replica in _replicate_iterable(_value.items(),
                                       count=count):
        yield dict(replica)


def replicator(count: int) -> _t.Callable[[_T], _t.Iterable[_T]]:
    """
    Returns function that replicates passed object.

    >>> triplicate = replicator(3)
    >>> list(map(tuple, triplicate(range(5))))
    [(0, 1, 2, 3, 4), (0, 1, 2, 3, 4), (0, 1, 2, 3, 4)]
    """
    return _functools.partial(replicate,
                              count=count)


duplicate = replicator(2)
duplicate.__doc__ = 'Duplicates given object.'



import functools as _functools
import itertools as _itertools
import typing as _t

from . import left as _left
from ._core.right import Applier as _Applier
from .functional import (compose as _compose,
                         flip as _flip)
from .iterating import expand as _expand
from .reversal import reverse as _reverse

_T1 = _t.TypeVar('_T1')
_T2 = _t.TypeVar('_T2')


def accumulator(
        _function: _t.Callable[[_T1, _T2], _T2], _initial: _T2
) -> _t.Callable[[_t.Iterable[_T1]], _t.Iterable[_t.Iterable[_T2]]]:
    """
    Returns function that yields cumulative results of given binary function
    starting from given initial object in direction from right to left.

    >>> def to_next_fraction(partial_denominator: int,
    ...                      reciprocal: float) -> float:
    ...     return partial_denominator + 1 / reciprocal
    >>> to_simple_continued_fractions = accumulator(to_next_fraction, 1)
    >>> from itertools import repeat
    >>> [round(fraction, 4)
    ...  for fraction in to_simple_continued_fractions(list(repeat(1, 10)))]
    [1, 2.0, 1.5, 1.6667, 1.6, 1.625, 1.6154, 1.619, 1.6176, 1.6182, 1.618]
    """
    return _t.cast(_t.Callable[[_t.Iterable[_T1]],
                               _t.Iterable[_t.Iterable[_T2]]],
                   _compose(_left.accumulator(_flip(_function), _initial),
                            _reverse))


def accumulate(_function: _t.Callable[[_T1, _T2], _T2],
               _initial: _T2,
               _iterable: _t.Iterable[_T1]) -> _t.Iterable[_T2]:
    """
    Yields cumulative results of given binary function
    starting from given initial object in direction from right to left.

    >>> def to_next_fraction(partial_denominator: int,
    ...                      reciprocal: float) -> float:
    ...     return partial_denominator + 1 / reciprocal
    >>> from itertools import repeat
    >>> [round(fraction, 4)
    ...  for fraction in accumulate(to_next_fraction, 1, list(repeat(1, 10)))]
    [1, 2.0, 1.5, 1.6667, 1.6, 1.625, 1.6154, 1.619, 1.6176, 1.6182, 1.618]
    """
    return _left.accumulate(_flip(_function), _initial, _reverse(_iterable))


def attacher(_value: _T1) -> _t.Callable[[_t.Iterable[_T1]], _t.Iterable[_T1]]:
    """
    Returns function that appends given object to iterable.

    >>> attach_hundred = attacher(100)
    >>> list(attach_hundred(range(10)))
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 100]
    """
    return _t.cast(_t.Callable[[_t.Iterable[_T1]], _t.Iterable[_T1]],
                   _Applier(attach, _value))


@_functools.singledispatch
def attach(_iterable: _t.Iterable[_T1], _value: _T1) -> _t.Iterable[_T1]:
    """
    Appends given object to the iterable.
    """
    yield from _itertools.chain(_iterable, _expand(_value))


@attach.register(list)
def _(_iterable: _t.List[_T1], _value: _T1) -> _t.List[_T1]:
    """
    Appends given object to the list.
    """
    return _iterable + [_value]


@attach.register(tuple)
def _(_iterable: _t.Tuple[_T1, ...], _value: _T1) -> _t.Tuple[_T1, ...]:
    """
    Appends given object to the tuple.
    """
    return _iterable + (_value,)


def folder(_function: _t.Callable[[_T1, _T2], _T2],
           _initial: _T2) -> _t.Callable[[_t.Iterable[_T1]], _T2]:
    """
    Returns function that cumulatively applies given binary function
    starting from given initial object in direction from right to left.

    >>> to_sum_evaluation_order = folder('({} + {})'.format, 0)
    >>> to_sum_evaluation_order(range(1, 10))
    '(1 + (2 + (3 + (4 + (5 + (6 + (7 + (8 + (9 + 0)))))))))'
    """
    left_folder = _left.folder(_flip(_function), _initial)
    return _t.cast(_t.Callable[[_t.Iterable[_T1]], _T2],
                   _compose(left_folder, _reverse))


def fold(_function: _t.Callable[[_T2, _T1], _T2],
         _initial: _T2,
         _iterable: _t.Iterable[_T1]) -> _T2:
    """
    Cumulatively applies given binary function
    starting from given initial object in direction from left to right.

    >>> fold('({} + {})'.format, 0, range(1, 10))
    '(1 + (2 + (3 + (4 + (5 + (6 + (7 + (8 + (9 + 0)))))))))'
    """
    return _functools.reduce(_flip(_function), _reverse(_iterable), _initial)


def applier(_function: _t.Callable[..., _T2],
            *args: _t.Any,
            **kwargs: _t.Any) -> _t.Callable[..., _T2]:
    """
    Returns function that behaves like given function
    with given arguments partially applied.
    Given positional arguments will be added to the right end.

    >>> square = applier(pow, 2)
    >>> square(10)
    100
    """
    return _Applier(_function, *args, **kwargs)



import functools as _functools
import typing as _t
from collections import (abc as _abc,
                         deque as _deque)

_T = _t.TypeVar('_T')


@_t.overload
def transpose(
        _value: _t.Collection[_t.Iterable[_T]]
) -> _t.Iterable[_t.Collection[_T]]:
    pass


@_t.overload
def transpose(
        _value: _t.Iterator[_t.Collection[_T]]
) -> _t.Collection[_t.Iterable[_T]]:
    pass


def transpose(_value: _T) -> _t.Any:
    """
    Transposes given object.

    >>> list(map(tuple, transpose(zip(range(10), range(10, 20)))))
    [(0, 1, 2, 3, 4, 5, 6, 7, 8, 9), (10, 11, 12, 13, 14, 15, 16, 17, 18, 19)]
    """
    return _transpose(_value)


@_functools.singledispatch
def _transpose(_value: _t.Any) -> _t.Any:
    raise TypeError(type(_value))


@_transpose.register(_abc.Iterable)
def _(
        _value: _t.Iterable[_t.Collection[_T]]
) -> _t.Collection[_t.Iterable[_T]]:
    iterator = iter(_value)
    try:
        first_elements = next(iterator)
    except StopIteration:
        return ()
    queues = [_deque([element]) for element in first_elements]

    def coordinate(queue: _t.Deque[_T]) -> _t.Iterable[_T]:
        while True:
            if not queue:
                try:
                    elements = next(iterator)
                except StopIteration:
                    return
                for sub_queue, element in zip(queues, elements):
                    sub_queue.append(element)
            yield queue.popleft()

    return tuple(map(coordinate, queues))


@_transpose.register(_abc.Collection)
def _(
        _value: _t.Collection[_t.Iterable[_T]]
) -> _t.Iterable[_t.Collection[_T]]:
    yield from zip(*_value)



import typing as _t
from operator import (not_,
                      xor)

import typing_extensions as _te

from . import left as _left
from .functional import (cleave as _cleave,
                         compose as _compose)

_Params = _te.ParamSpec('_Params')
_T = _t.TypeVar('_T')


def conjoin(
        *predicates: _t.Callable[_Params, bool]
) -> _t.Callable[_Params, bool]:
    """
    Returns conjunction of given predicates.

    >>> is_valid_constant_identifier = conjoin(str.isupper, str.isidentifier)
    >>> is_valid_constant_identifier('SECOND_SECTION')
    True
    >>> is_valid_constant_identifier('2ND_SECTION')
    False
    """
    return _compose(all, _cleave(*predicates))


def disjoin(
        *predicates: _t.Callable[_Params, bool]
) -> _t.Callable[_Params, bool]:
    """
    Returns disjunction of given predicates.

    >>> alphabetic_or_numeric = disjoin(str.isalpha, str.isnumeric)
    >>> alphabetic_or_numeric('Hello')
    True
    >>> alphabetic_or_numeric('42')
    True
    >>> alphabetic_or_numeric('Hello42')
    False
    """
    return _compose(any, _cleave(*predicates))


def exclusive_disjoin(
        *predicates: _t.Callable[_Params, bool]
) -> _t.Callable[_Params, bool]:
    """
    Returns exclusive disjunction of given predicates.

    >>> from keyword import iskeyword
    >>> valid_object_name = exclusive_disjoin(str.isidentifier, iskeyword)
    >>> valid_object_name('valid_object_name')
    True
    >>> valid_object_name('_')
    True
    >>> valid_object_name('1')
    False
    >>> valid_object_name('lambda')
    False
    """
    return _compose(_left.folder(xor, False), _cleave(*predicates))


def negate(
        predicate: _t.Callable[_Params, bool]
) -> _t.Callable[_Params, bool]:
    """
    Returns negated version of given predicate.

    >>> false_like = negate(bool)
    >>> false_like([])
    True
    >>> false_like([0])
    False
    """
    return _compose(not_, predicate)



import functools
import typing as _t
from itertools import chain

from paradigm.base import (OptionalParameter,
                           OverloadedSignature,
                           ParameterKind,
                           PlainSignature,
                           RequiredParameter)
from reprit import seekers
from reprit.base import generate_repr
from typing_extensions import final

from lz._core.functional import ApplierBase
from lz._core.signatures import (Parameter,
                                 Signature,
                                 plain_signature_to_parameters_by_kind,
                                 to_signature)

_Arg = _t.TypeVar('_Arg')
_KwArg = _t.TypeVar('_KwArg')
_Result = _t.TypeVar('_Result')
_T = _t.TypeVar('_T')


@final
class Applier(ApplierBase[_Arg, _KwArg, _Result]):
    def __init__(self,
                 function: _t.Callable[..., _Result],
                 *args: _Arg,
                 **kwargs: _KwArg) -> None:
        super().__init__(function, *args, **kwargs)

    def __call__(self, *args: _Arg, **kwargs: _KwArg) -> _Result:
        return self.function(*args, *self.args, **self.kwargs, **kwargs)

    __repr__ = generate_repr(__init__,
                             field_seeker=seekers.complex_)


@to_signature.register(Applier)
def _(value: Applier[_Arg, _KwArg, _Result]) -> Signature:
    return _bind_positionals_to_applier(
            to_signature(value.function).bind(**value.kwargs), value.args
    )


@functools.singledispatch
def _bind_positionals_to_applier(signature: Signature,
                                 args: _t.Tuple[_T, ...]) -> Signature:
    raise TypeError('Unsupported signature type: {type}.'
                    .format(type=type(signature)))


@_bind_positionals_to_applier.register(PlainSignature)
def _(signature: PlainSignature, args: _t.Tuple[_T, ...]) -> Signature:
    if not args:
        return signature
    parameters_by_kind = plain_signature_to_parameters_by_kind(signature)
    variadic_positionals = parameters_by_kind[
        ParameterKind.VARIADIC_POSITIONAL
    ]
    positionals = (parameters_by_kind[ParameterKind.POSITIONAL_ONLY]
                   + parameters_by_kind[ParameterKind.POSITIONAL_OR_KEYWORD])
    if len(args) > len(positionals) and not variadic_positionals:
        raise TypeError(f'Takes {len(positionals)} positional '
                        f'argument{"s" * (len(positionals) != 1)}, '
                        f'but {len(args)} '
                        f'{"was" if len(args) == 1 else "were"} given.')
    non_positionals = (parameters_by_kind[ParameterKind.KEYWORD_ONLY]
                       + parameters_by_kind[ParameterKind.VARIADIC_KEYWORD])
    signatures_parameters: _t.List[_t.Sequence[Parameter]] = []
    if len(args) <= len(positionals):
        signatures_parameters.append(positionals[:-len(args)]
                                     + non_positionals)
    if variadic_positionals:
        signatures_parameters.append(signature.parameters)
        for limit in range(1, len(args)):
            signatures_parameters.append(positionals[:-limit]
                                         + variadic_positionals
                                         + non_positionals)
    positionals_or_keywords = parameters_by_kind[
        ParameterKind.POSITIONAL_OR_KEYWORD
    ]
    optional_positionals_or_keywords = sum(
            map(OptionalParameter.__instancecheck__, positionals_or_keywords)
    )
    for offset in range(1, optional_positionals_or_keywords + 1):
        signatures_parameters.append(
                positionals[:-(len(args) + offset)]
                + [(OptionalParameter(annotation=parameter.annotation,
                                      **({'default': parameter.default}
                                         if parameter.has_default
                                         else {}),
                                      name=parameter.name,
                                      kind=ParameterKind.KEYWORD_ONLY)
                    if isinstance(parameter, OptionalParameter)
                    else RequiredParameter(annotation=parameter.annotation,
                                           kind=ParameterKind.KEYWORD_ONLY,
                                           name=parameter.name))
                   for parameter in positionals_or_keywords[-offset:]]
                + non_positionals
        )
    sub_signatures = [PlainSignature(*parameters,
                                     returns=signature.returns)
                      for parameters in signatures_parameters]
    return (sub_signatures[0]
            if len(sub_signatures) == 1
            else OverloadedSignature(*sub_signatures))


def flatten_signature(signature: Signature) -> _t.Iterable[PlainSignature]:
    if isinstance(signature, OverloadedSignature):
        yield from signature.signatures
    else:
        yield signature


@_bind_positionals_to_applier.register(OverloadedSignature)
def _(signature: OverloadedSignature, args: _t.Tuple[_T, ...]) -> Signature:
    sub_signatures = [_bind_positionals_to_applier(sub_signature, args)
                      for sub_signature in signature.signatures
                      if sub_signature.expects(*args)]
    if not sub_signatures:
        raise TypeError('No corresponding signature found.')
    flattened_sub_signatures = list(chain.from_iterable(map(flatten_signature,
                                                            sub_signatures)))
    return (flattened_sub_signatures[0]
            if len(flattened_sub_signatures) == 1
            else OverloadedSignature(*flattened_sub_signatures))



from functools import partial as _partial
import typing as _t


def instance_of(*types: type) -> _t.Callable[[_t.Any], bool]:
    """
    Creates predicate that checks if object is instance of given types.

    >>> is_any_string = instance_of(str, bytes, bytearray)
    >>> is_any_string(b'')
    True
    >>> is_any_string('')
    True
    >>> is_any_string(1)
    False
    """
    non_types = [candidate
                 for candidate in types
                 if not isinstance(candidate, type)]
    if non_types:
        raise TypeError(non_types)
    return _partial(_is_instance_of,
                    types=types)


def subclass_of(*types: type) -> _t.Callable[[type], bool]:
    """
    Creates predicate that checks if type is subclass of given types.

    >>> is_metaclass = subclass_of(type)
    >>> is_metaclass(type)
    True
    >>> is_metaclass(object)
    False
    """
    non_types = [candidate
                 for candidate in types
                 if not isinstance(candidate, type)]
    if non_types:
        raise TypeError(non_types)
    return _partial(_is_subclass_of,
                    types=types)


def _is_instance_of(value: _t.Any, types: _t.Tuple[type, ...]) -> bool:
    return isinstance(value, types)


def _is_subclass_of(value: type, types: _t.Tuple[type, ...]) -> bool:
    return issubclass(value, types)



import functools as _functools
import inspect as _inspect
import itertools as _itertools
import types as _types
import typing as _t

import typing_extensions as _te

from ._core import functional as _functional
from ._core.signatures import to_signature as _to_signature

_Arg = _t.TypeVar('_Arg')
_KwArg = _t.TypeVar('_KwArg')
_Params = _te.ParamSpec('_Params')
_Result = _t.TypeVar('_Result')
_T = _t.TypeVar('_T')
_T2 = _t.TypeVar('_T2')
_T3 = _t.TypeVar('_T3')


def identity(_value: _T) -> _T:
    """
    Returns object itself.

    >>> identity(0)
    0
    """
    return _value


def cleave(
        *functions: _t.Callable[..., _T]
) -> _t.Callable[..., _t.Iterable[_T]]:
    """
    Returns function that separately applies
    given functions to the same arguments.

    >>> to_min_and_max = cleave(min, max)
    >>> list(to_min_and_max(range(10)))
    [0, 9]
    >>> list(to_min_and_max(range(0), default=None))
    [None, None]
    """
    caller_frame_info = _inspect.stack()[1]
    return _functional.Cleavage(*functions,
                                file_path=caller_frame_info.filename,
                                line_number=caller_frame_info.lineno,
                                line_offset=0)


def combine(
        *maps: _t.Callable[[_T], _T2]
) -> _t.Callable[..., _t.Tuple[_T2, ...]]:
    """
    Returns function that applies each map to corresponding argument.

    >>> encoder_decoder = combine(str.encode, bytes.decode)
    >>> encoder_decoder('hello', b'world')
    (b'hello', 'world')
    """
    caller_frame_info = _inspect.stack()[1]
    return _functional.Combination(*maps,
                                   file_path=caller_frame_info.filename,
                                   line_number=caller_frame_info.lineno,
                                   line_offset=0)


def compose(
        _last_function: _t.Callable[[_T2], _T3],
        _penult_function: _t.Callable[..., _T2],
        *_rest_functions: _t.Callable[..., _t.Any]
) -> _t.Callable[_Params, _T3]:
    """
    Returns functions composition.

    >>> sum_of_first_n_natural_numbers = compose(sum, range)
    >>> sum_of_first_n_natural_numbers(10)
    45
    """
    caller_frame_info = _inspect.stack()[1]
    return _t.cast(
            _t.Callable[_Params, _T3],
            _functional.Composition(_last_function, _penult_function,
                                    *_rest_functions,
                                    file_path=caller_frame_info.filename,
                                    line_number=caller_frame_info.lineno,
                                    line_offset=0)
    )


def curry(
        _function: _t.Callable[..., _T2]
) -> _functional.Curry[_Arg, _KwArg, _T2]:
    """
    Returns curried version of given function.

    >>> curried_pow = curry(pow)
    >>> two_to_power = curried_pow(2)
    >>> two_to_power(10)
    1024
    """
    return _functional.Curry(_function, _to_signature(_function))


def pack(_function: _t.Callable[_Params, _T2]) -> _t.Callable[[_T, _T2], _T2]:
    """
    Returns function that works with single iterable parameter
    by unpacking elements to given function.

    >>> packed_int = pack(int)
    >>> packed_int(['10'])
    10
    >>> packed_int(['10'], {'base': 2})
    2
    """
    return _functools.partial(call, _function)


def call(_function: _t.Callable[_Params, _T2],
         _args: _Params.args,
         _kwargs: _Params.kwargs = _types.MappingProxyType({})) -> _T2:
    """
    Calls given function with given positional and keyword arguments.
    """
    return _function(*_args, **_kwargs)


def to_constant(_value: _T) -> _t.Callable[..., _T]:
    """
    Returns function that always returns given value.

    >>> always_zero = to_constant(0)
    >>> always_zero()
    0
    >>> always_zero(1)
    0
    >>> always_zero(how_about=2)
    0
    """
    return _functional.Constant(_value)


def flip(_function: _t.Callable[..., _T2]) -> _t.Callable[..., _T2]:
    """
    Returns function with positional arguments flipped.

    >>> flipped_power = flip(pow)
    >>> flipped_power(2, 4)
    16
    """
    return _t.cast(_t.Callable[..., _T2],
                   _functional.Flip.from_function(_function))


def flatmap(_function: _t.Callable[[_T], _t.Iterable[_T2]],
            *iterables: _t.Iterable[_T]) -> _t.Iterable[_T2]:
    """
    Applies given function to the arguments aggregated from given iterables
    and concatenates results into plain iterable.

    >>> list(flatmap(range, range(5)))
    [0, 0, 1, 0, 1, 2, 0, 1, 2, 3]
    """
    yield from _itertools.chain.from_iterable(map(_function, *iterables))



import functools as _functools
import itertools as _itertools
import typing as _t

from ._core.left import Applier as _Applier
from .iterating import expand as _expand

_T1 = _t.TypeVar('_T1')
_T2 = _t.TypeVar('_T2')


def accumulator(
        _function: _t.Callable[[_T2, _T1], _T2], _initial: _T2
) -> _t.Callable[[_t.Iterable[_T1]], _t.Iterable[_T2]]:
    """
    Returns function that yields cumulative results of given binary function
    starting from given initial object in direction from left to right.

    >>> import math
    >>> to_pi_approximations = accumulator(round, math.pi)
    >>> list(to_pi_approximations(range(5, 0, -1)))
    [3.141592653589793, 3.14159, 3.1416, 3.142, 3.14, 3.1]
    """
    return _functools.partial(accumulate, _function, _initial)


def accumulate(_function: _t.Callable[[_T2, _T1], _T2],
               _initial: _T2,
               _iterable: _t.Iterable[_T1]) -> _t.Iterable[_T2]:
    """
    Yields cumulative results of given binary function
    starting from given initial object in direction from left to right.

    >>> import math
    >>> list(accumulate(round, math.pi, range(5, 0, -1)))
    [3.141592653589793, 3.14159, 3.1416, 3.142, 3.14, 3.1]
    """
    yield from _itertools.accumulate(
            attach(_iterable, _initial),
            _t.cast(_t.Callable[[_t.Any, _t.Any], _T2], _function)
    )


def attacher(_value: _T1) -> _t.Callable[[_t.Iterable[_T1]], _t.Iterable[_T1]]:
    """
    Returns function that prepends given object to iterable.

    >>> attach_hundred = attacher(100)
    >>> list(attach_hundred(range(10)))
    [100, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    """
    return _functools.partial(attach,
                              _value=_value)


@_functools.singledispatch
def attach(_target: _t.Iterable[_T1], _value: _T1) -> _t.Iterable[_T1]:
    """
    Prepends given value to the target.
    """
    yield from _itertools.chain(_expand(_value), _target)


@attach.register(list)
def _(_target: _t.List[_T1], _value: _T1) -> _t.List[_T1]:
    """
    Prepends given object to the list.
    """
    return [_value] + _target


@attach.register(tuple)
def _(_target: _t.Tuple[_T1, ...], _value: _T1) -> _t.Tuple[_T1, ...]:
    """
    Prepends given value to the tuple.
    """
    return (_value,) + _target


def folder(_function: _t.Callable[[_T2, _T1], _T2],
           _initial: _T2) -> _t.Callable[[_t.Iterable[_T1]], _T2]:
    """
    Returns function that cumulatively applies given binary function
    starting from given initial object in direction from left to right.

    >>> to_sum_evaluation_order = folder('({} + {})'.format, 0)
    >>> to_sum_evaluation_order(range(1, 10))
    '(((((((((0 + 1) + 2) + 3) + 4) + 5) + 6) + 7) + 8) + 9)'
    """
    return _functools.partial(fold, _function, _initial)


def fold(_function: _t.Callable[[_T2, _T1], _T2],
         _initial: _T2,
         _iterable: _t.Iterable[_T1]) -> _T2:
    """
    Cumulatively applies given binary function
    starting from given initial object in direction from left to right.

    >>> fold('({} + {})'.format, 0, range(1, 10))
    '(((((((((0 + 1) + 2) + 3) + 4) + 5) + 6) + 7) + 8) + 9)'
    """
    return _functools.reduce(_function, _iterable, _initial)


def applier(_function: _t.Callable[..., _T2],
            *args: _T1,
            **kwargs: _T1) -> _t.Callable[..., _T2]:
    """
    Returns function that behaves like given function
    with given arguments partially applied.
    Given positional arguments will be added to the left end.

    >>> count_from_zero_to = applier(range, 0)
    >>> list(count_from_zero_to(10))
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    """
    return _Applier(_function, *args, **kwargs)



import collections as _collections
import functools as _functools
import itertools as _itertools
import typing as _t
from operator import is_not as _is_not

from .functional import flatmap as _flatmap

_T = _t.TypeVar('_T')
_T2 = _t.TypeVar('_T2')


@_functools.singledispatch
def capacity(_value: _t.Any) -> int:
    """
    Returns number of elements in value.

    >>> capacity(range(0))
    0
    >>> capacity(range(10))
    10
    """
    raise TypeError(type(_value))


@capacity.register(_collections.abc.Iterable)
def _(_iterable: _t.Iterable[_t.Any]) -> int:
    counter = _itertools.count()
    # order matters: if `counter` goes first,
    # then it will be incremented even for empty `iterable`
    _collections.deque(zip(_iterable, counter),
                       maxlen=0)
    return next(counter)


@capacity.register(_collections.abc.Collection)
def _(_iterable: _t.Sized) -> int:
    """
    Returns number of elements in sized iterable.
    """
    return len(_iterable)


def first(_iterable: _t.Iterable[_T]) -> _T:
    """
    Returns first element of iterable.

    >>> first(range(10))
    0
    """
    try:
        return next(iter(_iterable))
    except StopIteration as error:
        raise ValueError('Argument supposed to be non-empty.') from error


def last(_iterable: _t.Iterable[_T]) -> _T:
    """
    Returns last element of iterable.

    >>> last(range(10))
    9
    """
    try:
        return _collections.deque(_iterable,
                                  maxlen=1)[0]
    except IndexError as error:
        raise ValueError('Argument supposed to be non-empty.') from error


def cut(_iterable: _t.Iterable[_T],
        *,
        slice_: slice) -> _t.Iterable[_T]:
    """
    Selects elements from iterable based on given slice.

    Slice fields supposed to be unset or non-negative
    since it is hard to evaluate negative indices/step for arbitrary iterable
    which may be potentially infinite
    or change previous elements if iterating made backwards.
    """
    yield from _itertools.islice(_iterable,
                                 slice_.start, slice_.stop, slice_.step)


def cutter(_slice: slice) -> _t.Callable[[_t.Iterable[_T]], _t.Iterable[_T]]:
    """
    Returns function that selects elements from iterable based on given slice.

    >>> to_first_triplet = cutter(slice(3))
    >>> list(to_first_triplet(range(10)))
    [0, 1, 2]

    >>> to_second_triplet = cutter(slice(3, 6))
    >>> list(to_second_triplet(range(10)))
    [3, 4, 5]

    >>> cut_out_every_third = cutter(slice(0, None, 3))
    >>> list(cut_out_every_third(range(10)))
    [0, 3, 6, 9]
    """
    result = _functools.partial(cut,
                                slice_=_slice)
    result.__doc__ = ('Selects elements from iterable {slice}.'
                      .format(slice=_slice_to_description(_slice)))
    return result


def _slice_to_description(_slice: slice) -> str:
    """Generates human readable representation of `slice` object."""
    slice_description_parts = []
    start_is_specified = bool(_slice.start)
    if start_is_specified:
        slice_description_parts.append('starting from position {start}'
                                       .format(start=_slice.start))
    step_is_specified = _slice.step is not None
    if step_is_specified:
        slice_description_parts.append('with step {step}'
                                       .format(step=_slice.step))
    if _slice.stop is not None:
        stop_description_part = ('stopping at position {stop}'
                                 .format(stop=_slice.stop))
        if start_is_specified or step_is_specified:
            stop_description_part = 'and ' + stop_description_part
        slice_description_parts.append(stop_description_part)
    return ' '.join(slice_description_parts)


def chopper(
        _size: int
) -> _t.Callable[[_t.Iterable[_T]], _t.Iterable[_t.Sequence[_T]]]:
    """
    Returns function that splits iterable into chunks of given size.

    >>> in_three = chopper(3)
    >>> list(map(tuple, in_three(range(10))))
    [(0, 1, 2), (3, 4, 5), (6, 7, 8), (9,)]
    """
    return _functools.partial(chop,
                              size=_size)


@_functools.singledispatch
def chop(_iterable: _t.Iterable[_T],
         *,
         size: int) -> _t.Iterable[_t.Sequence[_T]]:
    """
    Splits iterable into chunks of given size.
    """
    iterator = iter(_iterable)
    yield from iter(lambda: tuple(_itertools.islice(iterator, size)), ())


@chop.register(_collections.abc.Sequence)
def _(_iterable: _t.Sequence[_T],
      *,
      size: int) -> _t.Iterable[_t.Sequence[_T]]:
    """
    Splits sequence into chunks of given size.
    """
    if not size:
        return
    for start in range(0, len(_iterable), size):
        yield _iterable[start:start + size]


# deque do not support slice notation
chop.register(_collections.deque, chop.registry[object])

in_two = chopper(2)
in_three = chopper(3)
in_four = chopper(4)


def slide(_iterable: _t.Iterable[_T],
          *,
          size: int) -> _t.Iterable[_t.Tuple[_T, ...]]:
    """
    Slides over iterable with window of given size.
    """
    iterator = iter(_iterable)
    initial = tuple(_itertools.islice(iterator, size))

    def shift(previous: _t.Tuple[_T, ...],
              element: _T) -> _t.Tuple[_T, ...]:
        return previous[1:] + (element,)

    yield from _itertools.accumulate(
            _itertools.chain([initial], iterator),
            _t.cast(_t.Callable[[_t.Any, _t.Any], _t.Tuple[_T, ...]], shift)
    )


def slider(_size: int) -> _t.Callable[[_t.Iterable[_T]],
_t.Iterable[_t.Tuple[_T, ...]]]:
    """
    Returns function that slides over iterable with window of given size.

    >>> pairwise = slider(2)
    >>> list(pairwise(range(10)))
    [(0, 1), (1, 2), (2, 3), (3, 4), (4, 5), (5, 6), (6, 7), (7, 8), (8, 9)]
    """
    return _functools.partial(slide,
                              size=_size)


pairwise = slider(2)
triplewise = slider(3)
quadruplewise = slider(4)


def header(_size: int) -> _t.Callable[[_t.Iterable[_T]], _t.Iterable[_T]]:
    """
    Returns function that selects elements from the beginning of iterable.
    Resulted iterable will have size not greater than given one.

    >>> to_first_pair = header(2)
    >>> list(to_first_pair(range(10)))
    [0, 1]
    """
    return cutter(slice(_size))


@_functools.singledispatch
def trail(_iterable: _t.Iterable[_T],
          *,
          size: int) -> _t.Iterable[_T]:
    """
    Selects elements from the end of iterable.
    Resulted iterable will have size not greater than given one.
    """
    return _collections.deque(_iterable,
                              maxlen=size)


@trail.register(_collections.abc.Sequence)
def _(iterable: _t.Sequence[_T],
      *,
      size: int) -> _t.Sequence[_T]:
    """
    Selects elements from the end of sequence.
    Resulted sequence will have size not greater than given one.
    """
    return iterable[-size:] if size else iterable[:size]


# deque do not support slice notation
trail.register(_collections.deque, trail.registry[object])


def trailer(_size: int) -> _t.Callable[[_t.Iterable[_T]], _t.Iterable[_T]]:
    """
    Returns function that selects elements from the end of iterable.
    Resulted iterable will have size not greater than given one.

    >>> to_last_pair = trailer(2)
    >>> list(to_last_pair(range(10)))
    [8, 9]
    """
    return _functools.partial(trail,
                              size=_size)


def mapper(
        _map: _t.Callable[[_T], _T2]
) -> _t.Callable[[_t.Iterable[_T]], _t.Iterable[_T2]]:
    """
    Returns function that applies given map to the each element of iterable.

    >>> to_str = mapper(str)
    >>> list(to_str(range(10)))
    ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']
    """
    return _t.cast(_t.Callable[[_t.Iterable[_T]], _t.Iterable[_T2]],
                   _functools.partial(map, _map))


def flatmapper(
        _map: _t.Callable[[_T], _t.Iterable[_T2]]
) -> _t.Callable[[_t.Iterable[_T]], _t.Iterable[_T2]]:
    """
    Returns function that applies map to the each element of iterable
    and flattens results.

    >>> relay = flatmapper(range)
    >>> list(relay(range(5)))
    [0, 0, 1, 0, 1, 2, 0, 1, 2, 3]
    """
    return _functools.partial(_flatmap, _map)


Group = _t.Tuple[_t.Hashable, _t.Iterable[_T]]


def group_by(_iterable: _t.Iterable[_T],
             *,
             key: _t.Callable[[_T], _t.Hashable]) -> _t.Iterable[Group[_T]]:
    """
    Groups iterable elements based on given key.
    """
    groups: _t.Dict[_t.Hashable, _t.List[_T]] = {}
    for element in _iterable:
        groups.setdefault(key(element), []).append(element)
    yield from groups.items()


def grouper(
        _key: _t.Callable[[_T], _t.Hashable]
) -> _t.Callable[[_t.Iterable[_T]], _t.Iterable[Group[_T]]]:
    """
    Returns function that groups iterable elements based on given key.

    >>> group_by_absolute_value = grouper(abs)
    >>> list(group_by_absolute_value(range(-5, 5)))
    [(5, [-5]), (4, [-4, 4]), (3, [-3, 3]), (2, [-2, 2]), (1, [-1, 1]), (0, [0])]

    >>> def modulo_two(number: int) -> int:
    ...     return number % 2
    >>> group_by_evenness = grouper(modulo_two)
    >>> list(group_by_evenness(range(10)))
    [(0, [0, 2, 4, 6, 8]), (1, [1, 3, 5, 7, 9])]
    """
    return _functools.partial(group_by,
                              key=_key)


def expand(_value: _T) -> _t.Iterable[_T]:
    """
    Wraps value into iterable.

    >>> list(expand(0))
    [0]
    """
    yield _value


def flatten(_iterable: _t.Iterable[_t.Iterable[_T]]) -> _t.Iterable[_T]:
    """
    Returns plain iterable from iterable of iterables.

    >>> list(flatten([range(5), range(10, 20)]))
    [0, 1, 2, 3, 4, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19]
    """
    yield from _itertools.chain.from_iterable(_iterable)


def interleave(_iterable: _t.Iterable[_t.Iterable[_T]]) -> _t.Iterable[_T]:
    """
    Interleaves elements from given iterable of iterables.

    >>> list(interleave([range(5), range(10, 20)]))
    [0, 10, 1, 11, 2, 12, 3, 13, 4, 14, 15, 16, 17, 18, 19]
    """
    iterators = _itertools.cycle(_t.cast(_t.Iterable[_t.Iterator[_T]],
                                         map(iter, _iterable)))
    while True:
        try:
            for iterator in iterators:
                yield next(iterator)
        except StopIteration:
            is_not_exhausted = _functools.partial(_is_not, iterator)
            iterators = _itertools.cycle(_itertools.takewhile(is_not_exhausted,
                                                              iterators))
        else:
            return



import functools
from typing import (Any,
                    Dict,
                    List,
                    Union)

from paradigm.base import (OptionalParameter,
                           OverloadedSignature,
                           ParameterKind,
                           PlainSignature,
                           RequiredParameter,
                           signature_from_callable)

Signature = Union[PlainSignature, OverloadedSignature]
Parameter = Union[OptionalParameter, RequiredParameter]


@functools.singledispatch
def to_signature(value: Any) -> Signature:
    return signature_from_callable(value)


def plain_signature_to_parameters_by_kind(
        signature: PlainSignature
) -> Dict[ParameterKind, List[Parameter]]:
    result: Dict[ParameterKind, List[Parameter]] = {kind: []
                                                    for kind in ParameterKind}
    for parameter in signature.parameters:
        result[parameter.kind].append(parameter)
    return result
